// Chat relayer for use with an 'iserve' communication strategy.
// Author: Jeff Dalton, with modifications by Stephen Potter
// Last modified 2 Jul 2010 by SP
// Copyright (C) 2009-2010, AIAI, The University of Edinburgh

// Set if avatars should be asked for permission to log their chat.
integer ASK_PERMISSION = TRUE;

// Link message commands.
// This is just a convention used with the I-X comm script.   The command
// is passed as the 2nd parameter (the integer) when llMessageLinked is called.

// Tells the comm script to send a message.
integer SEND_MESSAGE = 1701;

// Tells scripts (including this one) that we've received a message.
integer RECEIVED_MESSAGE = 1702;

integer OUTPUT_STATUS = 1703;

string PERMISSION_DIALOG_MESSAGE =
    "Do you give permission for your chat to be logged?";

string COMMAND_DIALOG_MESSAGE =
    "What would you like the Chat Relay to do?";

// What our dialogs might say:
string YES = "yes";
string NO = "no";
string DONT_LOG_ME = "don't log me";
string LOG_ME = "log my chat";
string STATUS = "status";
string CLEAR = "clear names";

// Chat channel used by this script's dialogs.
integer dialog_channel;

// Lists of avatar names to remember whether they've given permission or not.
list said_yes = [];
list said_no = [];

// List of messages waiting a 'yes' or 'no'.  Has the form:
//    [name message name message ...]
// New messages are added to the end.
list waiting = [];

// The name of the agent we'll send chat to.
// string chat_service = "chat-service";

send_chat(string sender_name, string message) {
    // Ask the send script to send it to the comm server.
    // which will forward it to the chat server.
    llMessageLinked(LINK_THIS, SEND_MESSAGE, sender_name + ": " + message, "");
}

process_waiting_messages(string target_name, integer send) {
    // Makes a new version of the 'waiting' list, removing all messages
    // from the target name, optinally sending them as they are processed.
    list revised_waiting = [];
    integer len = (waiting != []);
    integer i = 0;
    for (; i < len; i += 2) {
        string name = llList2String(waiting, i);
        string message = llList2String(waiting, i+1);
        if (name == target_name) {
            if (send) {
                send_chat(name, message);
            }
        }
        else {
            // Keep only entries whose name != the target.
            revised_waiting += [name, message];
        }
    }
    waiting = revised_waiting;
}

list waiting_names() {
    list result = [];
    integer len = (waiting != []);
    integer i = 0;
    for (; i < len; i += 2) {
        string name = llList2String(waiting, i);
        result += name;
    }
    return result;
}

integer is_avatar_id(key id) {
    return llGetOwnerKey(id) == id;
}

string avatar_permitted_logging(string name) {
    // Returns YES, NO, or "" if the avatar has not yet said.
    list name_as_list = [name];
    if (llListFindList(said_yes, name_as_list) >= 0) {
        return YES;
    }
    else if (llListFindList(said_no, name_as_list) >= 0) {
        return NO;
    }
    else {
        return "";
    }
}

list permission_change_answers(string avatar_name) {
    list result = [];
    string permitted = avatar_permitted_logging(avatar_name);
    if (permitted == YES) {
        return [DONT_LOG_ME];
    }
    else if (permitted == NO) {
        return [LOG_ME];
    }
    else {
        return [DONT_LOG_ME, LOG_ME];
    }
}

ensure_clear_permission_record(string name) {
    // This is called to ensure that a name doesn't end up on both said_yes
    // and said_no.  We have to worry about that because we might put up more
    // than one dialog for a user before we get an answer.   (We're not told
    // if the user clicks "Ignore", so we don't know whether a dialog is still
    // up nor not.)  However, we try to ask only questions that would change
    // the user's permission status, so that the user's reply indicates what
    // changes are needed; hence, it's only the possibility of multiple dialogs
    // that makes the precaution represented by this function necessary.
    said_yes = ensure_not_in(said_yes, name);
    said_no = ensure_not_in(said_no, name);
}
                
list ensure_not_in(list l, string target_name) {
    // Avaoid making a target-name-free copy of the list unless the
    // current version does contain the target name.
    list target = [target_name];
    if (llListFindList(l, target) < 0) {
        return l;
    }
    list result = [];
    integer i = (l != []) -1; // length - 1
    for (; i >= 0; --i) {
        string name = llList2String(l, i);
        if (name != target_name) {
            result = [name] + result; // building from end
        }
    }
    return result;
}


integer private_channel() {
    // Returns a channel number between -1073741823 (0xBFFFFFFF) and
    // -2147483648 (0x80000000) computed from the prim's key.
    // See http://wiki.secondlife.com/wiki/LlDialog
    return ((integer)("0x"+llGetSubString((string)llGetKey(),-8,-1)) & 0x3FFFFFFF)
           ^ 0xBFFFFFFF;
}


default
{
    state_entry()
    {
        llSetText("Chat Relay", <1,1,1>, 0.5);
        llListen(0, "", NULL_KEY, "" );
        dialog_channel = private_channel();
        llOwnerSay("Channel " + (string)dialog_channel);
        // When there were only two things dialogues might say:
        // llListen(dialog_channel, "", NULL_KEY, YES);
        // llListen(dialog_channel, "", NULL_KEY, NO);
        llListen(dialog_channel, "", NULL_KEY, "");
    }
    
    on_rez(integer start_param) {
        // Every time we're rezzed, reset the script.
        // This ensures that all local variables are set afresh.
        llResetScript();
    }
    
    listen(integer channel, string name, key id, string message )
    {
        if (channel == 0)
        {
            // We've heard some local chat.
        
            if (!ASK_PERMISSION) {
                send_chat(name, message);
                return;
            }
            
            // See whether we have permission to log it.
            string permitted = avatar_permitted_logging(name);
            
            if (permitted == YES) {
                // We have permission to log, so send the message to
                // the outside world.
                send_chat(name, message);
            }
            else if (permitted == NO) {
                // Denied permission to log, so ignore.
            }
            else if (! is_avatar_id(id) ) {
                // It's not an avatar, so assume we can log.
                send_chat(name, message);
            }
            else {
                // We have to ask permission.
                // First, save the message in the list of waiting messages.
                waiting += [name, message];
                // Ask the avatar.
                llDialog(id, PERMISSION_DIALOG_MESSAGE, [YES, NO], dialog_channel);
            }
        }
        else if (channel == dialog_channel)
        {
            // We have a reply from a permission or other dialog.
            
            if (message == YES) {
                // Remember that the avatar said YES.
                ensure_clear_permission_record(name);
                said_yes += name;
                // Send any waiting messages from the avatar.
                process_waiting_messages(name, TRUE);
            }
            else if (message == NO) {
                // Remember that the avatar said NO.
                ensure_clear_permission_record(name);
                said_no += name;
                // Delete any waiting messages from the avatar.
                process_waiting_messages(name, FALSE);
            }
            else if (message == LOG_ME) {
                ensure_clear_permission_record(name);
                said_yes += name;
                // Send any waiting messages from the avatar.
                process_waiting_messages(name, TRUE);
            }
            else if (message == DONT_LOG_ME) {
                ensure_clear_permission_record(name);
                said_no += name;
                // Delete any waiting messages from the avatar.
                process_waiting_messages(name, FALSE);
            }
            else if (message == STATUS) {
                integer n_waiting = (waiting != []) / 2;
                string reply = (string)n_waiting + " message(s) waiting.";
                if (said_yes != []) {
                    reply += "\nLogging: " + llList2CSV(said_yes);
                }
                if (said_no != []) {
                    reply += "\nNot logging: " + llList2CSV(said_no);
                }
                if (n_waiting > 0) {
                    reply += "\nHaven't replied: " +llList2CSV(waiting_names());
                }
                llInstantMessage(id, reply);
                llMessageLinked(LINK_THIS,OUTPUT_STATUS,"",id);
            }
            else if (message == CLEAR) {
                llResetScript();
            }
            
        }
 
    }
    
    link_message(integer sender_num, integer num, string message, key id)
    {
        if (num == RECEIVED_MESSAGE) {
            // A script has given us something to say.
            llSay(0, message);
        }
    }
    
    touch_start(integer n_touching)
    {
        integer i = 0;
        for (; i < n_touching; ++i) {
            string name = llDetectedName(i);
            key id = llDetectedKey(i);
            if (llDetectedGroup(i)) {
                // Avatar is in the same group as this object.
                llDialog(id, COMMAND_DIALOG_MESSAGE,
                         permission_change_answers(name) + [CLEAR, STATUS],
                         dialog_channel);
            }
            else {
                // Avatar's in some other group.
                llDialog(id, COMMAND_DIALOG_MESSAGE,
                         permission_change_answers(name),
                         dialog_channel);
            }
        }
    }
    
}
